"""
A framework and supporting libraries
for working with Flask, Connexion, and OpenAPI.

For usage, see :ref:`ligare-web`.

For examples and guides, see :ref:`examples-web-applications`.
"""

__version__: str = "0.7.0"



"""
Flask-specific integrations for :ref:`Ligare.web`.
"""

import re
import uuid
from logging import Logger
from typing import Awaitable, Callable, Dict, TypeAlias, TypeVar
from uuid import uuid4

from connexion import FlaskApp
from flask import Flask, Request, Response, request
from flask.typing import ResponseReturnValue
from injector import inject
from Ligare.web.middleware.context import get_trace_id

from ...config import Config
from ..consts import (
    CONTENT_SECURITY_POLICY_HEADER,
    CORRELATION_ID_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER,
    CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER,
    HOST_HEADER,
    INCOMING_REQUEST_MESSAGE,
    ORIGIN_HEADER,
    OUTGOING_RESPONSE_MESSAGE,
    REQUEST_COOKIE_HEADER,
    RESPONSE_COOKIE_HEADER,
)

# pyright: reportUnusedFunction=false


# Fixes type problems when using @inject with @app.before_request and @app.after_request.
# The main difference with these types as opposed to the Flask-defined types is that
# these types allow the handler to take any arguments, versus no arguments or just Response.
AfterRequestCallable: TypeAlias = (
    Callable[..., Response] | Callable[..., Awaitable[Response]]
)
BeforeRequestCallable: TypeAlias = (
    Callable[..., ResponseReturnValue | None]
    | Callable[..., Awaitable[ResponseReturnValue | None]]
)
T_request_callable = TypeVar(
    "T_request_callable", bound=BeforeRequestCallable | AfterRequestCallable | None
)
# Fixes type problems when using Flask-Injector, which automatically sets up any bound
# error handlers. The error handler types in Flask have the same problem as After/BeforeRequestCallable
# in that their parameters are `[Any]` rather than `...`. This relaxes that. The type
# is valid here because Flask-Injector does actually include the provided types as parameters.
ErrorHandlerCallable = (
    Callable[..., ResponseReturnValue] | Callable[..., Awaitable[ResponseReturnValue]]
)
T_error_handler = TypeVar("T_error_handler", bound=ErrorHandlerCallable)

TFlaskApp = Flask | FlaskApp
T_flask_app = TypeVar("T_flask_app", bound=TFlaskApp)


def bind_requesthandler(
    app: T_flask_app,
    flask_app_handler: Callable[
        [T_flask_app, Callable[..., Response | None]], T_request_callable
    ],
):
    def wrapper(request_callable: Callable[..., Response | None]) -> T_request_callable:
        return flask_app_handler(app, request_callable)

    return wrapper


def _get_correlation_id(log: Logger) -> str:
    correlation_id = get_trace_id().CorrelationId

    if not correlation_id:
        correlation_id = _get_correlation_id_from_headers(log)

    return correlation_id


def _get_correlation_id_from_headers(log: Logger) -> str:
    try:
        correlation_id = request.headers.get(CORRELATION_ID_HEADER)

        if correlation_id:
            # validate format
            _ = uuid.UUID(correlation_id)
        else:
            correlation_id = str(uuid4())
            log.info(
                f'Generated new UUID "{correlation_id}" for {CORRELATION_ID_HEADER} request header.'
            )

        return correlation_id

    except ValueError as e:
        log.warning(f"Badly formatted {CORRELATION_ID_HEADER} received in request.")
        raise e


@inject
def _log_all_api_requests(
    request: Request,
    config: Config,
    log: Logger,
):
    request_headers_safe: dict[str, str] = dict(request.headers)

    correlation_id = _get_correlation_id(log)

    if (
        request_headers_safe.get(REQUEST_COOKIE_HEADER)
        and config.flask
        and config.flask.session
    ):
        request_headers_safe[REQUEST_COOKIE_HEADER] = re.sub(
            rf"({config.flask.session.cookie.name}=)[^;]+(;|$)",
            r"\1<redacted>\2",
            request_headers_safe[REQUEST_COOKIE_HEADER],
        )

    log.info(
        INCOMING_REQUEST_MESSAGE,
        request.method,
        request.url,
        request.host,
        request.remote_addr,
        request.remote_user,
        extra={
            "props": {
                "correlation_id": correlation_id,
                "headers": request_headers_safe,
            }
        },
    )


@inject
def _ordered_api_response_handers(response: Response, config: Config, log: Logger):
    _wrap_all_api_responses(response, config, log)
    _log_all_api_responses(response, config, log)
    return response


def _wrap_all_api_responses(response: Response, config: Config, log: Logger):
    correlation_id = _get_correlation_id(log)

    cors_domains: list[str] | None = None
    if config.web.security.cors.origins:
        cors_domains = config.web.security.cors.origins
    else:
        if not response.headers.get(CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER):
            cors_domains = (
                (header_value := request.headers.get(ORIGIN_HEADER))
                and [header_value]
                or None
            )
            if not cors_domains:
                cors_domains = (
                    (header_value := request.headers.get(HOST_HEADER))
                    and [header_value]
                    or None
                )

    if cors_domains:
        # although the config allows multiple, the header may only contain one.
        # either only one will have been set, or we take the first from the config.
        # this may not be valid in all cases, and so the ASGI CORSMiddleware sould
        # be used instead
        response.headers[CORS_ACCESS_CONTROL_ALLOW_ORIGIN_HEADER] = cors_domains[0]

    response.headers[CORS_ACCESS_CONTROL_ALLOW_CREDENTIALS_HEADER] = str(
        config.web.security.cors.allow_credentials
    )

    response.headers[CORS_ACCESS_CONTROL_ALLOW_METHODS_HEADER] = ",".join(
        config.web.security.cors.allow_methods
    )

    response.headers[CORRELATION_ID_HEADER] = correlation_id

    if config.web.security.csp:
        response.headers[CONTENT_SECURITY_POLICY_HEADER] = config.web.security.csp

    # if config.flask and config.flask.openapi and config.flask.openapi.use_swagger:
    #    # Use a permissive CSP for the Swagger UI
    #    # https://github.com/swagger-api/swagger-ui/issues/7540
    #    FIXME what to use other than `request` w/ Connexion middleware?
    #    if request.path.startswith("/ui/") or (
    #        request.url_rule and request.url_rule.endpoint == "/v1./v1_swagger_ui_index"
    #    ):
    #        response.headers[
    #            CONTENT_SECURITY_POLICY_HEADER
    #        ] = "default-src 'self'; style-src 'self' 'unsafe-inline'; img-src 'self'; script-src 'self' 'unsafe-inline'"


def _log_all_api_responses(response: Response, config: Config, log: Logger):
    correlation_id = _get_correlation_id(log)

    response_headers_safe: Dict[str, str] = dict(response.headers)

    if (
        response_headers_safe.get(RESPONSE_COOKIE_HEADER)
        and config.flask
        and config.flask.session
    ):
        response_headers_safe[RESPONSE_COOKIE_HEADER] = re.sub(
            rf"({config.flask.session.cookie.name}=)[^;]+(;|$)",
            r"\1<redacted>\2",
            response_headers_safe[RESPONSE_COOKIE_HEADER],
        )

    log.info(
        OUTGOING_RESPONSE_MESSAGE,
        response.status_code,
        response.status,
        extra={
            "props": {
                "correlation_id": correlation_id,
                "headers": response_headers_safe,
            }
        },
    )


def register_flask_api_request_handlers(app: Flask):
    return bind_requesthandler(app, Flask.before_request)(_log_all_api_requests)


def register_flask_api_response_handlers(app: Flask):
    # TODO consider moving request/response logging to the WSGI app
    # apparently Flask may not call this if unhandled exceptions occur
    return bind_requesthandler(app, Flask.after_request)(_ordered_api_response_handers)



import json
from enum import Enum

from flask import Response


class OutputType(Enum):
    PNG = "png"
    TIFF = "tiff"


def create_attachment_response(
    data: bytes,
    attachment_filename: str = "download",
    output_type: OutputType = OutputType.PNG,
) -> Response:
    """
    Create a new `Response` with appropriate headers for sending
    PNG data. The response body is the PNG data.

    :param bytes data: The PNG data.
    :param str attachment_filename: The name of the filename in the
      `Content-Disposition` response header, defaults to "download".
    :return Response: The PNG HTTP response.
    """
    resp = Response(data)
    resp.headers["Content-Disposition"] = (
        f"attachment; filename={attachment_filename}.{output_type.value}"
    )
    resp.headers["Content-Type"] = f"image/{output_type.value}"
    return resp


def create_BadRequest_response(error_msg: str) -> Response:
    """
    Create a 400 `Response.

    :param str error_msg: An error message to include in the response's JSON data.
    :return Response: A 400 Bed Request HTTP response. The response body is
      a JSON object with the following structure:
      {
        "error_msg": error_msg,
        "status": "Bad Request",
        "status_code": 400
      }
    """
    resp = Response(
        json.dumps({
            "error_msg": error_msg,
            "status": "Bad Request",
            "status_code": 400,
        }),
        400,
    )
    # 400 might not be accurate
    resp.headers["Content-Type"] = "application/json"
    return resp



import pytest
from flask import Flask
from Ligare.web.application import ApplicationBuilder, configure_openapi
from Ligare.web.config import Config, FlaskConfig, FlaskOpenApiConfig
from Ligare.web.testing.create_app import CreateOpenAPIApp
from mock import MagicMock
from pytest_mock import MockerFixture


class TestCreateOpenAPIApp(CreateOpenAPIApp):
    def test__CreateOpenAPIApp__configure_openapi__requires_flask_config(self):
        with pytest.raises(
            Exception,
            match=r"^OpenAPI configuration is empty\. Review the `openapi` section of your application's `config\.toml`\.$",
        ):
            _ = configure_openapi(Config())

    def test__CreateOpenAPIApp__configure_openapi__creates_flask_app_using_config(
        self, mocker: MockerFixture
    ):
        connexion_mock = mocker.patch("Ligare.web.application.FlaskApp")

        app_name = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__configure_openapi__creates_flask_app_using_config.__name__}-app_name"
        spec_path = ".."

        _ = configure_openapi(
            Config(
                flask=FlaskConfig(
                    app_name=app_name,
                    openapi=FlaskOpenApiConfig(spec_path=spec_path),
                )
            )
        )

        connexion_mock.assert_called()
        assert connexion_mock.call_args.kwargs
        assert connexion_mock.call_args.kwargs["specification_dir"] == spec_path

    def test__CreateOpenAPIApp__create_app__loads_config_from_toml(
        self, basic_config: Config, mocker: MockerFixture
    ):
        _ = mocker.patch(
            "Ligare.AWS.ssm.SSMParameters",
            return_value=MagicMock(load_config=MagicMock(return_value=None)),
        )
        load_config_mock = mocker.patch(
            "Ligare.programming.application.load_config", return_value=basic_config
        )

        toml_filename = f"{TestCreateOpenAPIApp.test__CreateOpenAPIApp__create_app__loads_config_from_toml.__name__}-config.toml"
        application_builder = ApplicationBuilder(Flask).use_configuration(
            lambda config_builder: config_builder.with_config_filename(toml_filename)
        )
        _ = application_builder.build()
        assert load_config_mock.called
        assert load_config_mock.call_args and load_config_mock.call_args[0]
        assert load_config_mock.call_args[0][1] == toml_filename



"""
:ref:`Ligare.web`'s integration with :ref:`Ligare.platform.feature_flag`.
"""

from dataclasses import dataclass
from functools import wraps
from logging import Logger
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    ParamSpec,
    Sequence,
    TypedDict,
    TypeVar,
    cast,
)

from connexion import FlaskApp, request
from flask import Blueprint, Flask, abort
from injector import Binder, Injector, Module, inject, provider, singleton
from Ligare.database.dependency_injection import ScopedSessionModule
from Ligare.database.types import MetaBase
from Ligare.platform.feature_flag.caching_feature_flag_router import (
    CachingFeatureFlagRouter,
)
from Ligare.platform.feature_flag.caching_feature_flag_router import (
    FeatureFlag as CachingFeatureFlag,
)
from Ligare.platform.feature_flag.db_feature_flag_router import DBFeatureFlagRouter
from Ligare.platform.feature_flag.db_feature_flag_router import (
    FeatureFlag as DBFeatureFlag,
)
from Ligare.platform.feature_flag.db_feature_flag_router import FeatureFlagTableBase
from Ligare.platform.feature_flag.feature_flag_router import (
    FeatureFlag,
    FeatureFlagRouter,
    TFeatureFlag,
)
from Ligare.programming.config import AbstractConfig
from Ligare.programming.patterns.dependency_injection import ConfigurableModule
from Ligare.web.middleware.sso import login_required
from pydantic import BaseModel
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import override


class FeatureFlagConfig(BaseModel):
    api_base_url: str = "/platform"
    access_role_name: str | bool | None = None


class Config(AbstractConfig):
    @override
    def post_load(self) -> None:
        return super().post_load()

    feature_flag: FeatureFlagConfig


class FeatureFlagPatchRequest(TypedDict):
    name: str
    enabled: bool


@dataclass
class FeatureFlagPatch:
    name: str
    enabled: bool


class FeatureFlagRouterModule(ConfigurableModule, Generic[TFeatureFlag]):
    """
    An Injector module to provide an instance of a Feature Flag Router.

    :type ConfigurableModule: Enables loading of Ligare.web.middleware.feature_flags.Config config information
    :type Generic[TFeatureFlag]: The type of Feature Flag containing feature flag data
    """

    def __init__(
        self, t_feature_flag: type[FeatureFlagRouter[TFeatureFlag]] | None = None
    ) -> None:
        """
        Get an instance of the generic FeatureFlagRouterModule.

        :param type[FeatureFlagRouter[TFeatureFlag]] | None t_feature_flag: The type of Feature Flag Router this Injector module provides, defaults to None
        """
        self._t_feature_flag = type(self) if t_feature_flag is None else t_feature_flag
        super().__init__()

    @override
    @staticmethod
    def get_config_type() -> type[AbstractConfig]:
        """
        Get the Feature Flag Router Module configuration type.

        :return type[Ligare.web.middleware.feature_flags.Config]:
        """
        return Config

    @singleton
    @provider
    def _provide_feature_flag_router(
        self, injector: Injector
    ) -> FeatureFlagRouter[FeatureFlag]:
        """
        Provide the instance of FeatureFlagRouter[TFeatureFlag].

        :param Injector injector: The IoC container
        :return FeatureFlagRouter[FeatureFlag]: The FeatureFlagRouter[TFeatureFlag] instance
        """
        return cast(FeatureFlagRouter[FeatureFlag], injector.get(self._t_feature_flag))


class DBFeatureFlagRouterModule:
    def __new__(
        cls,
        feature_flag_table: type[FeatureFlagTableBase[MetaBase]],
        bases: list[MetaBase | type[MetaBase]] | None = None,
    ):
        """
        An Injector module to work with Feature Flags stored in a database.

        :param type[FeatureFlagTableBase] feature_flag_table: The base table type representing columns the database table contains, and other aspects of SQLAlchemy Table object
        :param list[MetaBase  |  type[MetaBase]] | None bases: All MetaBase types an SQLAlchemy ScopedSession instance is responsible for, defaults to None
        :return _type_: _description_
        """

        class _DBFeatureFlagRouterModule(FeatureFlagRouterModule[DBFeatureFlag]):
            _feature_flag_table: type[FeatureFlagTableBase[MetaBase]] = (
                feature_flag_table
            )
            _bases: list[MetaBase | type[MetaBase]] | None = bases

            def __init__(self) -> None:
                super().__init__(DBFeatureFlagRouter)

            @override
            def configure(self, binder: Binder) -> None:
                binder.install(ScopedSessionModule(self._bases))

            @singleton
            @provider
            def _provide_db_feature_flag_router(
                self, injector: Injector
            ) -> FeatureFlagRouter[DBFeatureFlag]:
                """
                Provide an instance of the FeatureFlagRouter for the DBFeatureFlag type.
                """
                return cast(
                    FeatureFlagRouter[DBFeatureFlag], injector.get(self._t_feature_flag)
                )

            @singleton
            @provider
            def _provide_db_feature_flag_router_table_base(
                self,
            ) -> type[FeatureFlagTableBase[MetaBase]]:
                """
                Provide the base table type of the Feature Flag table.
                """
                return self._feature_flag_table

        return _DBFeatureFlagRouterModule


class CachingFeatureFlagRouterModule(FeatureFlagRouterModule[CachingFeatureFlag]):
    """
    An Injector module to work with Feature Flags in-memory.
    """

    def __init__(self) -> None:
        super().__init__(CachingFeatureFlagRouter)

    @singleton
    @provider
    def _provide_caching_feature_flag_router(
        self, injector: Injector
    ) -> FeatureFlagRouter[CachingFeatureFlag]:
        """
        Provide an instance of the FeatureFlagRouter for the CachingFeatureFlag type.
        """
        return cast(
            FeatureFlagRouter[CachingFeatureFlag], injector.get(self._t_feature_flag)
        )


P = ParamSpec("P")
R = TypeVar("R")


@inject
def _get_feature_flag_blueprint(app: Flask, config: FeatureFlagConfig, log: Logger):
    """
    Create and return a Flask Blueprint with API endpoints for querying and managing Feature Flags.

    :param Flask app: The registered Flask application this Blueprint will be applied to.
    :param FeatureFlagConfig config: The Feature Flag configuration type.
    :param Logger log:
    :raises ValueError:
    :return Blueprint:
    """
    feature_flag_blueprint = Blueprint(
        "feature_flag", __name__, url_prefix=f"{config.api_base_url}"
    )

    access_role = config.access_role_name

    def _login_required(require_flask_login: bool):
        """
        Decorate an API endpoint with the correct flask_login authentication
        method given the requirements of the API endpoint.

        require_flask_login is ignored if flask_login has been configured.

        If flask_login has _not_ been configured:
            * If require_flask_login is True, a warning is logged and the request is aborted with a 405 error
            * If require_flask_login is False, the endpoint function is executed

        :param bool require_flask_login: Determine whether flask_login must be configured for this endpoint to function
        :return _type_: _description_
        """

        def __login_required(
            fn: Callable[P, R],
        ) -> Callable[P, R]:
            authorization_implementation: Callable[..., Any]

            if access_role is False:
                authorization_implementation = fn
            # None means no roles were specified, but a session is still required
            elif access_role is None or access_role is True:
                authorization_implementation = login_required(fn)
            else:
                authorization_implementation = login_required([access_role])(fn)

            @wraps(fn)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                if not hasattr(app, "login_manager"):
                    if require_flask_login:
                        log.warning(
                            "The Feature Flag module expects flask_login to be configured in order to control access to feature flag modifications. flask_login has not been configured, so the Feature Flag modification API is disabled."
                        )

                        return abort(405)
                    else:
                        return fn(*args, **kwargs)
                return authorization_implementation(*args, **kwargs)

            return wrapper

        return __login_required

    @feature_flag_blueprint.route("/feature_flag", methods=("GET",))
    @_login_required(False)
    @inject
    def feature_flag(  # pyright: ignore[reportUnusedFunction]
        feature_flag_router: FeatureFlagRouter[FeatureFlag],
    ) -> dict[str, Any] | tuple[dict[str, Any], Literal[404]]:
        """
        Query Feature Flag values.
        This API does not require a Flask session.

        :param FeatureFlagRouter[FeatureFlag] feature_flag_router: The Feature Flag router type.
        :raises ValueError:
        :return dict[str, Any] | tuple[dict[str, Any], Literal[404]]:
        """
        request_query_names: list[str] | None = request.query_params.getlist("name")

        feature_flags: Sequence[FeatureFlag]
        missing_flags: set[str] | None = None
        if request_query_names is None or not request_query_names:
            feature_flags = feature_flag_router.get_feature_flags()
        elif isinstance(request_query_names, list):  # pyright: ignore[reportUnnecessaryIsInstance]
            feature_flags = feature_flag_router.get_feature_flags(request_query_names)
            missing_flags = set(request_query_names).difference(
                set([feature_flag.name for feature_flag in feature_flags])
            )
        else:
            raise ValueError("Unexpected type from Flask query parameters.")

        response: dict[str, Any] = {}
        problems: list[Any] = []

        if missing_flags is not None:
            for missing_flag in missing_flags:
                problems.append({
                    "title": "feature flag not found",
                    "detail": "Queried feature flag does not exist.",
                    "instance": missing_flag,
                    "status": 404,
                    "type": None,
                })
            response["problems"] = problems

        elif not feature_flags:
            problems.append({
                "title": "No feature flags found",
                "detail": "Queried feature flags do not exist.",
                "instance": "",
                "status": 404,
                "type": None,
            })
            response["problems"] = problems

        if feature_flags:
            response["data"] = feature_flags
            return response
        else:
            return response, 404

    @feature_flag_blueprint.route("/feature_flag", methods=("PATCH",))
    @_login_required(True)
    @inject
    async def feature_flag_patch(  # pyright: ignore[reportUnusedFunction]
        feature_flag_router: FeatureFlagRouter[FeatureFlag],
    ) -> dict[str, Any] | tuple[dict[str, Any], Literal[404]]:
        """
        Modify Feature Flag values.
        Requires a Flask session and appropriate User Role access. If access is not granted, HTTP status 405 is returned.

        :return dict[str, Any] | tuple[dict[str, Any], Literal[404]]:
        """
        feature_flags_request: list[FeatureFlagPatchRequest] = await request.json()

        feature_flags = [
            FeatureFlagPatch(name=flag["name"], enabled=flag["enabled"])
            for flag in feature_flags_request
        ]

        changes: list[Any] = []
        problems: list[Any] = []
        for flag in feature_flags:
            try:
                change = feature_flag_router.set_feature_is_enabled(
                    flag.name, flag.enabled
                )
                changes.append(change)
            except LookupError:
                problems.append({
                    "title": "feature flag not found",
                    "detail": "Feature flag to PATCH does not exist. It must be created first.",
                    "instance": flag.name,
                    "status": 404,
                    "type": None,
                })

        response: dict[str, Any] = {}

        if problems:
            response["problems"] = problems

        if changes:
            response["data"] = changes
            return response
        else:
            return response, 404

    return feature_flag_blueprint


class FeatureFlagMiddlewareModule(Module):
    """
    Enable the use of Feature Flags and a Feature Flag management API.
    """

    def __init__(
        self,
        feature_flag_router_module_type: type[FeatureFlagRouterModule[TFeatureFlag]],
    ) -> None:
        """
        Create the Injector module.

        :param type[FeatureFlagRouterModule[TFeatureFlag]] feature_flag_router_module_type:
        """
        super().__init__()
        self._feature_flag_router_module_type = feature_flag_router_module_type

    @override
    def configure(self, binder: Binder) -> None:
        binder.install(self._feature_flag_router_module_type())
        super().configure(binder)

    def register_middleware(self, app: FlaskApp):
        app.add_middleware(FeatureFlagMiddlewareModule.FeatureFlagMiddleware)

    class FeatureFlagMiddleware:
        """
        ASGI middleware for Feature Flags.

        This middleware create a Flask blueprint the enables a Feature Flag management API.
        """

        _app: ASGIApp

        def __init__(self, app: ASGIApp):
            super().__init__()
            self._app = app

        @inject
        async def __call__(
            self,
            scope: Scope,
            receive: Receive,
            send: Send,
            app: Flask,
            injector: Injector,
            log: Logger,
        ) -> None:
            async def wrapped_send(message: Any) -> None:
                # Only run during startup of the application
                if (
                    scope["type"] != "lifespan"
                    or message["type"] != "lifespan.startup.complete"
                    or not scope["app"]
                ):
                    return await send(message)

                log.debug("Registering FeatureFlag blueprint.")
                app.register_blueprint(
                    injector.call_with_injection(_get_feature_flag_blueprint)
                )
                log.debug("FeatureFlag blueprint registered.")

                return await send(message)

            await self._app(scope, receive, wrapped_send)



import uuid
from collections.abc import Collection
from contextvars import ContextVar
from logging import Logger
from typing import Any, Callable, Literal, NamedTuple, NewType, TypedDict, cast
from uuid import uuid4

from connexion import ConnexionMiddleware
from injector import inject
from Ligare.programming.collections.dict import AnyDict
from Ligare.web.middleware.consts import CORRELATION_ID_HEADER
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import final

MiddlewareRequestDict = TypedDict(
    "MiddlewareRequestDict",
    {
        "type": Literal["http"],
        "http_version": str,
        "method": Literal[
            "GET",
            "POST",
            "PATCH",
            "PUT",
            "DELETE",
            "OPTIONS",
            "HEAD",
            "CONNECT",
            "TRACE",
        ],
        "path": str,
        "raw_path": bytes,
        "root_path": str,
        "scheme": str,
        "query_string": bytes,
        "headers": list[tuple[bytes, bytes]],
        "client": list[
            str | int
        ],  # this is actually a 2-item list whose first item is a str and second item is an int
        "server": list[str | int],  # same here
        "extensions": dict[str, dict[str, str]],
        "state": AnyDict,
        "app": ConnexionMiddleware,
        "starlette.exception_handlers": tuple[
            dict[type[Any], Callable[..., Any]], dict[Any, Any]
        ],
        "path_params": dict[Any, Any],
    },
)

MiddlewareResponseDict = TypedDict(
    "MiddlewareResponseDict",
    {
        "type": Literal["http.response.start"],
        "status": int,
        "headers": list[tuple[bytes, bytes]],
    },
)


# copied from connexion.utils
def extract_content_type(
    headers: list[tuple[bytes, bytes]] | dict[str, str],
) -> str | None:
    """Extract the mime type and encoding from the content type headers.

    :param headers: Headers from ASGI scope

    :return: The content type if available in headers, otherwise None
    """
    content_type: str | None = None

    header_pairs: Collection[tuple[str | bytes, str | bytes]] = (
        headers.items() if isinstance(headers, dict) else headers
    )
    for key, value in header_pairs:
        # Headers can always be decoded using latin-1:
        # https://stackoverflow.com/a/27357138/4098821
        if isinstance(key, bytes):
            decoded_key: str = key.decode("latin-1")
        else:
            decoded_key = key

        if decoded_key.lower() == "content-type":
            if isinstance(value, bytes):
                content_type = value.decode("latin-1")
            else:
                content_type = value
            break

    return content_type


# copied from connexion.utils
def split_content_type(content_type: str | None) -> tuple[str | None, str | None]:
    """Split the content type in mime_type and encoding. Other parameters are ignored."""
    mime_type, encoding = None, None

    if content_type is None:
        return mime_type, encoding

    # Check for parameters
    if ";" in content_type:
        mime_type, parameters = content_type.split(";", maxsplit=1)

        # Find parameter describing the charset
        prefix = "charset="
        for parameter in parameters.split(";"):
            if parameter.startswith(prefix):
                encoding = parameter[len(prefix) :]
    else:
        mime_type = content_type
    return mime_type, encoding


RequestId = NewType("RequestId", str)
CorrelationId = NewType("CorrelationId", str)

REQUEST_ID_CTX_KEY = "requestId"
CORRELATION_ID_CTX_KEY = "correlationId"

_request_id_ctx_var: ContextVar[RequestId | None] = ContextVar(
    REQUEST_ID_CTX_KEY, default=None
)
_correlation_id_ctx_var: ContextVar[CorrelationId | None] = ContextVar(
    CORRELATION_ID_CTX_KEY, default=None
)


class TraceId(NamedTuple):
    CorrelationId: CorrelationId | None
    RequestId: RequestId | None


def get_trace_id() -> TraceId:
    return TraceId(_correlation_id_ctx_var.get(), _request_id_ctx_var.get())


@final
class CorrelationIdMiddleware:
    """
    Generate a Correlation ID for each request.

    https://github.com/encode/starlette/issues/420
    """

    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ["http", "websocket"]:
            await self.app(scope, receive, send)
            return

        correlation_id = _correlation_id_ctx_var.set(CorrelationId(str(uuid4())))

        await self.app(scope, receive, send)

        _correlation_id_ctx_var.reset(correlation_id)


@final
class RequestIdMiddleware:
    """
    Generate a Trace ID for each request.
    If X-Correlation-Id is set in the request headers, that ID is used instead.
    """

    _app: ASGIApp

    def __init__(self, app: ASGIApp):
        super().__init__()
        self._app = app

    @inject
    async def __call__(
        self, scope: Scope, receive: Receive, send: Send, log: Logger
    ) -> None:
        if scope["type"] not in ["http", "websocket"]:
            return await self._app(scope, receive, send)

        # extract the request ID from the request headers if it is set

        request = cast(MiddlewareRequestDict, scope)
        request_headers = request.get("headers")

        content_type = extract_content_type(request_headers)
        _, encoding = split_content_type(content_type)
        if encoding is None:
            encoding = "utf-8"

        try:
            request_id_header_encoded = CORRELATION_ID_HEADER.lower().encode(encoding)

            request_id: bytes | None = next(
                (
                    request_id
                    for (header, request_id) in request_headers
                    if header == request_id_header_encoded
                ),
                None,
            )

            if request_id:
                # validate format
                request_id_decoded = request_id.decode(encoding)
                _ = uuid.UUID(request_id_decoded)
                request_id_token = _request_id_ctx_var.set(
                    RequestId(request_id_decoded)
                )
            else:
                request_id_decoded = str(uuid4())
                request_id = request_id_decoded.encode(encoding)
                request_headers.append((
                    request_id_header_encoded,
                    request_id,
                ))
                request_id_token = _request_id_ctx_var.set(
                    RequestId(request_id_decoded)
                )
                log.info(
                    f'Generated new UUID "{request_id}" for {CORRELATION_ID_HEADER} request header.'
                )
        except ValueError as e:
            log.warning(f"Badly formatted {CORRELATION_ID_HEADER} received in request.")
            raise e

        async def wrapped_send(message: Any) -> None:
            nonlocal scope
            nonlocal send

            if message["type"] != "http.response.start":
                return await send(message)

            # include the request ID in response headers

            response = cast(MiddlewareResponseDict, message)
            response_headers = response["headers"]

            content_type = extract_content_type(response_headers)
            _, encoding = split_content_type(content_type)
            if encoding is None:
                encoding = "utf-8"

            response_headers.append((
                request_id_header_encoded,
                request_id,
            ))

            return await send(message)

        await self._app(scope, receive, wrapped_send)

        _request_id_ctx_var.reset(request_id_token)



"""
The framework API for creating Flask and Connexion applications.
"""

import logging
from collections import defaultdict
from dataclasses import dataclass
from os import path
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Protocol,
    TypeVar,
    cast,
    final,
    overload,
)

from connexion import FlaskApp
from connexion.options import SwaggerUIOptions
from flask import Blueprint, Flask
from flask_injector import FlaskInjector
from injector import Injector
from lib_programname import get_path_executed_script
from Ligare.AWS.ssm import SSMParameters
from Ligare.programming.application import ApplicationBase
from Ligare.programming.application import (
    ApplicationBuilder as GenericApplicationBuilder,
)
from Ligare.programming.application import AppModule
from Ligare.programming.collections.dict import NestedDict
from Ligare.programming.config import AbstractConfig, ConfigBuilder, load_config
from Ligare.programming.config.exceptions import ConfigInvalidError
from Ligare.programming.exception import BuilderBuildError, InvalidBuilderStateError
from typing_extensions import Self, override

from .config import Config, FlaskConfig
from .middleware import (
    register_api_request_handlers,
    register_api_response_handlers,
    register_context_middleware,
    register_error_handlers,
)
from .middleware.dependency_injection import configure_dependencies

_get_program_dir = lambda: path.dirname(get_path_executed_script())
_get_exec_dir = lambda: path.abspath(".")

TApp = Flask | FlaskApp
T_app = TypeVar("T_app", bound=TApp)

TAppConfig = TypeVar("TAppConfig", bound=Config)


@dataclass
class AppInjector(Generic[T_app]):
    """
    Contains an instantiated `T_app` application in `app`,
    and its associated `FlaskInjector` IoC container.

    :param T_app Generic: An instance of Flask or FlaskApp.
    :param flask_inject FlaskInject: The applications IoC container.
    """

    app: T_app
    flask_injector: FlaskInjector


@dataclass
class CreateAppResult(ApplicationBase, Generic[T_app]):
    """
    Contains an instantiated Flask application and its
    associated application "container." This is either
    the same Flask instance, or an OpenAPI application.

    :param flask_app Generic: The Flask application.
    :param app_injector AppInjector[T_app]: The application's wrapper and IoC container.
    """

    flask_app: Flask
    app_injector: AppInjector[T_app]

    @property
    def app(self) -> T_app:
        return self.app_injector.app

    @property
    def injector(self) -> Injector:
        return self.app_injector.flask_injector.injector

    @overload
    def run(self) -> None: ...

    @overload
    def run(
        self: "CreateAppResult[Flask]",
        *,
        host: str | None = None,
        port: int | None = None,
        debug: bool | None = None,
        load_dotenv: bool = True,
        **options: Any,
    ) -> None:
        """
        Call this method to start your application.
        This method is a passthrough for `Flask.run`.

        Reference https://github.com/encode/uvicorn/blob/fe3910083e3990695bc19c2ef671dd447262ae18/uvicorn/main.py#L463

        :param host: The hostname this application should accept requests for.
                              If `None`, the value in the application's `FlaskConfig` instance is used;
                              otherwise, this parameter value is used.
        :param port: The port this application should listen on for requests.
                              If `None`, the value in the application's `FlaskConfig` instance is used;
                              otherwise, this parameter value is used.
        """
        ...

    @overload
    def run(
        self: "CreateAppResult[FlaskApp]",
        *,
        import_string: str | None = None,
        host: str | None = None,
        # uvicorn's default is 8000 but we default to 5000
        # and try to load the value from a config file
        port: int | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Call this method to start your application.
        This method is partly a passthrough for `uvicorn.run`.

        Reference https://github.com/encode/uvicorn/blob/fe3910083e3990695bc19c2ef671dd447262ae18/uvicorn/main.py#L463

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param host: The hostname this application should accept requests for.
                              If `None`, the value in the application's `FlaskConfig` instance is used;
                              otherwise, this parameter value is used.
        :param port: The port this application should listen on for requests.
                              If `None`, the value in the application's `FlaskConfig` instance is used;
                              otherwise, this parameter value is used.
        """
        ...

    @override
    def run(
        self,
        *,
        import_string: str | None = None,
        # uvicorn's default is 127.0.0.1 but we default to localhost
        # and try to load the value from a config file
        host: str | None = None,
        # uvicorn's default is 8000 but we default to 5000
        # and try to load the value from a config file
        port: int | None = None,
        **kwargs: Any,
    ):
        """
        Call this method to start your application.
        This method is partly a passthrough for `uvicorn.run`.

        Reference https://github.com/encode/uvicorn/blob/fe3910083e3990695bc19c2ef671dd447262ae18/uvicorn/main.py#L463

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param host: The hostname this application should accept requests for.
                              If `None`, the value in the application's `FlaskConfig` instance is used;
                              otherwise, this parameter value is used.
        :param port: The port this application should listen on for requests.
                              If `None`, the value in the application's `FlaskConfig` instance is used;
                              otherwise, this parameter value is used.
        """
        app = self.app_injector.app
        injector = self.app_injector.flask_injector.injector
        config = injector.get(FlaskConfig)

        host = host or config.host
        port = port or int(config.port)

        if isinstance(app, FlaskApp):
            app.run(
                import_string=import_string,  # pyright: ignore[reportArgumentType] the connexion type annotation is wrong; `None` is supported.
                host=host,
                port=port,
                **kwargs,
            )
        else:
            app.run(
                host=host,
                port=port,
                **kwargs,
            )


FlaskAppResult = CreateAppResult[Flask]
OpenAPIAppResult = CreateAppResult[FlaskApp]


class UseConfigurationCallback(Protocol[TAppConfig]):
    """
    The callback for configuring an application's configuration.

    :param TAppConfig Protocol: The AbstractConfig type to be configured.
    """

    def __call__(
        self,
        config_builder: ConfigBuilder[TAppConfig],
        config_overrides: dict[str, Any],
    ) -> "None | ConfigBuilder[TAppConfig]":
        """
        Set up parameters for the application's configuration.

        :param ConfigBuilder[TAppConfig] config_builder: The ConfigBuilder instance.
        :param dict[str, Any] config_overrides: A dictionary of key/values that are applied over all keys that might exist in an instantiated config.
        :raises InvalidBuilderStateError: Upon a call to `build()`, the builder is misconfigured.
        :raises BuilderBuildError: Upon a call to `build()`, a failure occurred during the instantiation of the configuration.
        :raises Exception: Upon a call to `build()`, an unknown error occurred.
        :return None | ConfigBuilder[TAppConfig]: The callback may return `None` or the received `ConfigBuilder` instance so as to support the use of lambdas. This return value is not used.
        """


@final
class ApplicationConfigBuilder(Generic[TAppConfig]):
    _DEFAULT_CONFIG_FILENAME: str = "config.toml"

    def __init__(self) -> None:
        self._config_value_overrides: dict[str, Any] = {}
        self._config_builder: ConfigBuilder[TAppConfig] = ConfigBuilder[TAppConfig]()
        self._config_filename: str = ApplicationConfigBuilder._DEFAULT_CONFIG_FILENAME
        self._use_filename: bool = False
        self._use_ssm: bool = False

    def with_config_builder(self, config_builder: ConfigBuilder[TAppConfig]) -> Self:
        self._config_builder = config_builder
        return self

    def with_config_types(self, configs: list[type[AbstractConfig]] | None) -> Self:
        _ = self._config_builder.with_configs(configs)
        return self

    def with_config_type(self, config_type: type[AbstractConfig]) -> Self:
        _ = self._config_builder.with_config(config_type)
        return self

    def with_config_value_overrides(self, values: dict[str, Any]) -> Self:
        self._config_value_overrides = values
        return self

    def with_config_filename(self, filename: str) -> Self:
        self._config_filename = filename
        self._use_filename = True
        return self

    def enable_ssm(self, value: bool) -> Self:
        """
        Try to load config from AWS SSM. If `use_filename` was configured,
        a failed attempt to load from SSM will instead attempt to load from
        the configured filename. If `use_filename` is not configured and SSM
        fails, an exception is raised. If SSM succeeds, `build` will not
        load from the configured filename.

        :param bool value: Whether to use SSM
        :return Self:
        """
        self._use_ssm = value
        return self

    def build(self) -> TAppConfig | None:
        _ = self._config_builder.with_root_config(Config)  # pyright: ignore[reportArgumentType]

        if not (self._use_ssm or self._use_filename):
            raise InvalidBuilderStateError(
                f"Cannot build the application config without either `{ApplicationConfigBuilder[TAppConfig].enable_ssm.__name__}` or `{ApplicationConfigBuilder[TAppConfig].with_config_filename.__name__}` having been configured."
            )

        config_type = self._config_builder.build()

        full_config: TAppConfig | None = None
        SSM_FAIL_ERROR_MSG = "Unable to load configuration. SSM parameter load failed and the builder is configured not to load from a file."
        if self._use_ssm:
            try:
                # requires that aws-ssm.ini exists and is correctly configured
                ssm_parameters = SSMParameters()
                full_config = ssm_parameters.load_config(config_type)

                if not self._use_filename and full_config is None:
                    raise BuilderBuildError(SSM_FAIL_ERROR_MSG)
            except Exception as e:
                if self._use_filename:
                    logging.getLogger().info("SSM parameter load failed.", exc_info=e)
                else:
                    raise BuilderBuildError(SSM_FAIL_ERROR_MSG) from e

        if self._use_filename and full_config is None:
            if self._config_value_overrides:
                full_config = load_config(
                    config_type, self._config_filename, self._config_value_overrides
                )
            else:
                full_config = load_config(config_type, self._config_filename)

        return full_config


class ApplicationConfigBuilderCallback(Protocol[TAppConfig]):
    def __call__(
        self,
        config_builder: ApplicationConfigBuilder[TAppConfig],
    ) -> "None | ApplicationConfigBuilder[TAppConfig]":
        """
        A method used to configure an `ApplicationConfigBuilder`.
        Call the builder methods on `config_builder` to set the
        desired options.

        **Do not call `build()`** as it is called by the `ApplicationBuilder`.

        :param ApplicationConfigBuilder[TAppConfig] config_builder:
        :return None | ApplicationConfigBuilder[TAppConfig]: Any return value is ignored.
        """
        ...


@final
class ApplicationBuilder(GenericApplicationBuilder[T_app]):
    def __init__(
        self,
        exec: type[T_app] | Callable[..., T_app],
    ) -> None:
        super().__init__(exec=exec)

    def with_flask_app_name(self, value: str | None) -> Self:
        self._config_overrides["app_name"] = value
        return self

    def with_flask_env(self, value: str | None) -> Self:
        self._config_overrides["env"] = value
        return self

    @override
    def use_configuration(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        __application_config_builder_callback: ApplicationConfigBuilderCallback[Config],
    ) -> Self:
        """
        Execute changes to the builder's `ApplicationConfigBuilder[TAppConfig]` instance.

        `__builder_callback` can return `None`, or the instance of `ApplicationConfigBuilder[TAppConfig]` passed to its `config_builder` argument.
        This allowance is so lambdas can be used; `ApplicationBuilder[T_app, TAppConfig]` does not use the return value.
        """
        super().use_configuration(__application_config_builder_callback)  # pyright: ignore[reportCallIssue,reportArgumentType]
        return self

    @override
    def build(self) -> CreateAppResult[T_app]:
        config_overrides = cast(NestedDict[str, Any], defaultdict(dict))

        if (
            override_app_name := self._config_overrides.get("app_name", None)
        ) is not None and override_app_name != "":
            config_overrides["flask"]["app_name"] = override_app_name

        if (
            override_env := self._config_overrides.get("env", None)
        ) is not None and override_env != "":
            config_overrides["flask"]["env"] = override_env

        _ = self._application_config_builder.with_root_config_type(
            Config
        ).with_config_value_overrides(config_overrides)

        config = cast(Config, self._build_config())
        self._register_config_modules(config)

        if config.flask is None:
            raise ConfigInvalidError(
                "You must set [flask] in the application configuration. Review the documentation for the Ligare.web TOML format and requirements."
            )

        if not config.flask.app_name:
            raise ConfigInvalidError(
                "You must set the Flask application name in the [flask.app_name] config or FLASK_APP envvar. Review the documentation for the Ligare.web TOML format and requirements."
            )

        if not self._app_module_set:
            _ = self.with_module(
                AppModule(self._exec, None, import_name=config.flask.app_name)
            )

        if config.flask.openapi is not None:
            openapi = configure_openapi(config)
            app = cast(T_app, openapi)
        else:
            app = cast(T_app, configure_blueprint_routes(config))

        register_error_handlers(app)
        _ = register_api_request_handlers(app)
        _ = register_api_response_handlers(app)
        _ = register_context_middleware(app)

        modules = self._build_application_modules()

        flask_injector = configure_dependencies(app, application_modules=modules)

        flask_app = app.app if isinstance(app, FlaskApp) else app
        return CreateAppResult[T_app](
            flask_app, AppInjector[T_app](app, flask_injector)
        )


def configure_openapi(config: Config, name: Optional[str] = None):
    """
    Instantiate Connexion and set Flask logging options
    """

    if (
        config.flask is None
        or config.flask.openapi is None
        or config.flask.openapi.spec_path is None
    ):
        raise Exception(
            "OpenAPI configuration is empty. Review the `openapi` section of your application's `config.toml`."
        )

    if config.flask.openapi.use_swagger and config.flask.openapi.swagger_url in [
        "/",
        "",
    ]:
        raise Exception(
            f'The configured Swagger URL "{config.flask.openapi.swagger_url}" cannot be used. Remove `flask.openapi.swagger_url` from your configuration, or change the value to something else.'
        )

    exec_dir = _get_exec_dir()

    connexion_app = FlaskApp(
        config.flask.app_name,
        specification_dir=exec_dir,
        swagger_ui_options=SwaggerUIOptions(
            swagger_ui=config.flask.openapi.use_swagger,
            swagger_ui_path=config.flask.openapi.swagger_url
            or SwaggerUIOptions.swagger_ui_path,
        ),
    )
    app = connexion_app.app
    config.update_flask_config(app.config)

    _ = connexion_app.add_api(
        f"{config.flask.app_name}/{config.flask.openapi.spec_path}",
        validate_responses=config.flask.openapi.validate_responses,
    )

    # FIXME what's the new way to get this URL?
    # if config.flask.openapi.use_swagger:
    #    # App context needed for url_for.
    #    # This can only run after connexion is instantiated
    #    # because it registers the swagger UI url.
    #    with app.app_context():
    #        app.logger.info(
    #            f"Swagger UI can be accessed at {url_for('/./_swagger_ui_index', _external=True)}"
    #        )

    return connexion_app


def configure_blueprint_routes(
    config: Config, blueprint_import_subdir: str = "endpoints"
):
    """
    Register Flask blueprints and API routes
    """
    if config.flask is None:
        raise Exception(
            "Flask configuration is empty. Review the `flask` section of your application's `config.toml`."
        )

    app = Flask(config.flask.app_name)
    config.update_flask_config(app.config)

    blueprint_modules = _import_blueprint_modules(app, blueprint_import_subdir)
    _register_blueprint_modules(app, blueprint_modules)
    return app


def _import_blueprint_modules(app: Flask, blueprint_import_subdir: str):
    from importlib.util import module_from_spec, spec_from_file_location
    from pathlib import Path

    from flask import Blueprint

    program_dir = _get_program_dir()
    blueprint_import_dir = Path(program_dir, blueprint_import_subdir)

    module_paths = blueprint_import_dir.glob("*.py")

    blueprint_modules: list[Blueprint] = []

    for path in module_paths:
        if not (path.is_file() or path.name == "__init__.py"):
            continue
        # load the module from its path
        # and execute it
        spec = spec_from_file_location(path.name.rstrip(".py"), str(path))
        if spec is None or spec.loader is None:
            raise Exception(f"Module cannot be created from path {path}")
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        # find all Flask blueprints in
        # the module and register them
        for module_name, module_var in vars(module).items():
            if module_name.endswith("_blueprint") or isinstance(module_var, Blueprint):
                blueprint_modules.append(module_var)

    return blueprint_modules


def _register_blueprint_modules(app: Flask, blueprint_modules: list[Blueprint]):
    for module in blueprint_modules:
        app.register_blueprint(module)



"""
Middleware functionality for handing HTTP requests and responses.
"""

import json
from logging import Logger
from typing import Awaitable, Callable, TypeAlias, TypeVar

from connexion import FlaskApp
from flask import Flask, Response
from flask.typing import ResponseReturnValue
from Ligare.web.middleware.flask import (
    register_flask_api_request_handlers,
    register_flask_api_response_handlers,
)
from Ligare.web.middleware.openapi import (
    register_openapi_api_request_handlers,
    register_openapi_api_response_handlers,
    register_openapi_context_middleware,
)
from werkzeug.exceptions import HTTPException, Unauthorized

# pyright: reportUnusedFunction=false


# Fixes type problems when using @inject with @app.before_request and @app.after_request.
# The main difference with these types as opposed to the Flask-defined types is that
# these types allow the handler to take any arguments, versus no arguments or just Response.
AfterRequestCallable: TypeAlias = (
    Callable[..., Response] | Callable[..., Awaitable[Response]]
)
BeforeRequestCallable: TypeAlias = (
    Callable[..., ResponseReturnValue | None]
    | Callable[..., Awaitable[ResponseReturnValue | None]]
)
T_request_callable = TypeVar(
    "T_request_callable", bound=BeforeRequestCallable | AfterRequestCallable | None
)
# Fixes type problems when using Flask-Injector, which automatically sets up any bound
# error handlers. The error handler types in Flask have the same problem as After/BeforeRequestCallable
# in that their parameters are `[Any]` rather than `...`. This relaxes that. The type
# is valid here because Flask-Injector does actually include the provided types as parameters.
ErrorHandlerCallable = (
    Callable[..., ResponseReturnValue] | Callable[..., Awaitable[ResponseReturnValue]]
)
T_error_handler = TypeVar("T_error_handler", bound=ErrorHandlerCallable)

TFlaskApp = Flask | FlaskApp
T_flask_app = TypeVar("T_flask_app", bound=TFlaskApp)

RegisterMiddlewareCallback = Callable[[FlaskApp], None]


def bind_errorhandler(
    app: TFlaskApp,
    code_or_exception: type[Exception] | int,
) -> Callable[[T_error_handler], T_error_handler | None]:
    if isinstance(app, Flask):
        return app.errorhandler(code_or_exception)
    else:
        return app.app.errorhandler(code_or_exception)


def register_api_request_handlers(app: TFlaskApp):
    if isinstance(app, Flask):
        return register_flask_api_request_handlers(app)
    else:
        return register_openapi_api_request_handlers(app)


def register_api_response_handlers(app: TFlaskApp):
    if isinstance(app, Flask):
        # TODO consider moving request/response logging to the WSGI app
        # apparently Flask may not call this if unhandled exceptions occur
        return register_flask_api_response_handlers(app)
    else:
        return register_openapi_api_response_handlers(app)


def register_context_middleware(app: TFlaskApp):
    if not isinstance(app, FlaskApp):
        return
    return register_openapi_context_middleware(app)


def register_error_handlers(app: TFlaskApp):
    @bind_errorhandler(app, Exception)
    # @inject
    def catch_all_catastrophic(error: Exception, log: Logger):
        # error: connexion.lifecycle.ConnexionRequest, log: ZeroDivisionError
        log.exception(error)

        response = {
            "status_code": 500,
            "error_msg": "Unknown error.",
            "status": "Internal Server Error",
        }
        return response, 500

    @bind_errorhandler(app, HTTPException)
    # @inject
    def catch_all(error: HTTPException, log: Logger):
        log.exception(error)

        response = {
            "status_code": error.code,
            "error_msg": error.description,
            "status": error.name,
        }

        return (
            response,
            error.code if error.code is not None else 500,
        )

    @bind_errorhandler(app, 401)
    # @inject
    def unauthorized(error: Unauthorized, log: Logger):
        log.info(error)

        if error.response is None or not isinstance(error.response, Response):
            response = {
                "status_code": 401,
                "error_msg": error.description,
                "status": error.name,
            }
            return response, 401

        response = error.response
        data = {
            "status_code": response.status_code,
            "error_msg": response.data.decode(),
            "status": response.status,
        }
        response.data = json.dumps(data)
        return response



from flask import Flask
from flask.testing import FlaskClient
from Ligare.web.middleware.dependency_injection import configure_dependencies
from Ligare.web.testing.create_app import ClientInjector, CreateFlaskApp


class TestApplicationBuilder(CreateFlaskApp):
    def test__configure_dependencies__registers_correct_Flask_instance_with_AppModule(
        self, flask_client: ClientInjector[FlaskClient]
    ):
        flask_injector = configure_dependencies(flask_client.client.application)

        flask_app = flask_injector.injector.get(Flask)
        assert flask_app == flask_client.client.application
